/* VirtuinoSE example: Overview with text
 * Supported boards: ESP8266 & ESP32 
 * Created by Ilias Lamprou
 * Updated Oct 27 2018
 * More examples: http://virtuino.com/index.php/virtuino-se/20-virtuino-se-code-examples
 */
// ------ VirtuinoSE V Memory Table
//  V1 memory -> Pin D5 or 16 digital output Relay1
//  V2 memory -> Pin D6 or 17 digital output Relay2
//  V3 memory -> Pin D7 or 4  digital input Button1
//  V4 memory -> Pin A0  analog input
//  V5 memory -> Sensor 1 value 
//  V6 memory -> Sensor 2 value
//  V7 memory -> is a variable the value of which can be changed by VirtuinoSE
//  V8 memory -> Text


#ifdef ESP8266
 #include <ESP8266WiFi.h>
 #define V1_relay1_pin D5
 #define V2_relay2_pin D6
 #define V3_button1_pin D7
 #else
 #include <WiFi.h>
 #define V1_relay1_pin 16
 #define V2_relay2_pin 17
 #define V3_button1_pin 4
 #endif

int   V4_analog_pin = A0;
int   V5_sensor1 = 0;
int   V6_sensor2 = 0;
float V7_variable = 0;
String V8_text = "";



//------------------- USER SETTINGS ---------------------------------
const char* ssid = "WIFI NETWORK";           // WIFI network SSID
const char* password = "PASSWORD";          // WIFI network PASSWORD

WiFiServer wifiServer(8000);                  // Server port 

const char* ssid_AP = "Virtuino network";     // Access point network SSID
const char* password_AP = "1234567890";       // Access point network PASSWORD

#define VIRTUINO_KEY  "1234"                  // Virtuino KEY, only the requests that include this key will be acceptable
                                              // disable the line if you don't want to use a key
unsigned long timeStrored=0;

float V7_variable_stored=0;
//============================================================== connectToWiFiNetwork
void connectToWiFiNetwork(){
  Serial.println("Connecting to "+String(ssid));
  
   // If you don't want to config IP manually disable the next four lines
  IPAddress ip(192, 168, 1, 150);            // where 150 is the desired IP Address   
  IPAddress gateway(192, 168, 1, 1);         // set gateway to match your network
  IPAddress subnet(255, 255, 255, 0);        // set subnet mask to match your network
  WiFi.config(ip, gateway, subnet);          // If you don't want to config IP manually disable this line
 
  WiFi.mode(WIFI_STA);                       // Configure the module as station only.
   WiFi.begin(ssid, password);
   while (WiFi.status() != WL_CONNECTED) {
     delay(500);
     Serial.print(".");
    }
   Serial.println("");
   Serial.println("WiFi connected");
   Serial.println(WiFi.localIP());
}

//=================================================================
void initAccessPoint(){
  Serial.print("Setting soft-AP ... ");                   // Default IP: 192.168.4.1
   WiFi.mode(WIFI_AP);                                     // Configure the module as Access point only.  Set WiFi.mode(WIFI_AP_STA); to config module as Acces point and station
   boolean result = WiFi.softAP(ssid_AP,password_AP);      // set the Access point SSID and password
   if(result == true)  {
    Serial.println("Server Ready");
    Serial.println(WiFi.softAPIP());
   }
   else Serial.println("Failed!");
}

//============================================================== setup
//==============================================================
void setup() {
  Serial.begin(9600);
  while (!Serial) {;}
  
  //initAccessPoint();               //enable this line to use the esp as AP  
  connectToWiFiNetwork();      //enable this line to connect the module your wifi network  
  
  wifiServer.begin();

  pinMode(V1_relay1_pin, OUTPUT);            
  pinMode(V2_relay2_pin, OUTPUT);
  pinMode(V3_button1_pin, INPUT);
  }

//============================================================== onCommandReceived
//==============================================================
/* This function is called every time Virtuino app sends a request to server 
 * The 'variableIndex' is the V memory of Virtuino app
 * The 'valueAsText' is the value that has sent from the app
 * If the 'valueAsText" is ? you have to return the value that you want to be displayed on Virtuino panel
 */
 
String onCommandReceived(int variableIndex, String valueAsText){
   Serial.println("=== onCommandReceived  variableIndex="+String(variableIndex)+"   valueAsText="+valueAsText);

  //-------- Enable or disable pins
  //--- code for the V1_relay1_pin (read-write)
  if (variableIndex==1){            // Virtuino V1 memory = V1_relay1_pin
      if (valueAsText=="?") return String(digitalRead(V1_relay1_pin));       // return the state of pin V1_relay1_pin as text
      int v=valueAsText.toInt();
      if (v==1) digitalWrite(V1_relay1_pin,HIGH);          // change the state of pin D4 as text
        else if (v==0) digitalWrite(V1_relay1_pin,LOW);
   }

  //--- code for the V2_relay2_pin (read-write)
  else if (variableIndex==2){            // Virtuino V2 memory = V2_relay2_pin
      if (valueAsText=="?") return String(digitalRead(V2_relay2_pin));       // return the state of pin V1_relay1_pin as text
      int v=valueAsText.toInt();
      if (v==1) digitalWrite(V2_relay2_pin,HIGH);          // change the state of pin D4 as text
        else if (v==0) digitalWrite(V2_relay2_pin,LOW);
   }

  //--- code for the V3_button1_pin (read only)
  else if (variableIndex==3){            // Virtuino V3 memory = V3_button1_pin
      if (valueAsText=="?") return String(digitalRead(V3_button1_pin));       // return the state of pin V1_relay1_pin as text
   }

  //--- code for the V4_analog_pin (read only)
   else if (variableIndex==4){            // Virtuino V4 memory = V4_analog_pin
     if (valueAsText=="?") return String(V4_analog_pin);       
      }

//--- code for the V5_sensor1  (read only)
   else if (variableIndex==5){            // Virtuino V5 memory = V5_sensor1
     if (valueAsText=="?") return String(V5_sensor1);       
      }

//--- code for the V6_sensor2  (read only)
  else if (variableIndex==6){            // Virtuino V6 memory = V6_sensor2
     if (valueAsText=="?") return String(V6_sensor2);       
      }

//--- code for the V7_variable (read-write)
  else if (variableIndex==7){            // Virtuino V7 memory = V7_variable
      if (valueAsText=="?") return String(V7_variable);       // return the value of V7_variable
      else V7_variable=valueAsText.toFloat();
   }

//-----Read Write text (read-write)   
     else if (variableIndex==8) {                      // Virtuino V8 memory = V8_text
          if (valueAsText=="?") return urlencode(&V8_text);   // encode and decode the text 
          else  V8_text=urldecode(&valueAsText);
        }
        

  return "";
}


//============================================================== loop
//==============================================================
void loop() {
  
  virtuinoRun();

  
  
   if (millis()-timeStrored>5000) {      // read sensor value every 5 seconds. Avoid to use delay on your code
      V5_sensor1 = random(100);     // replace the random value with your sensors' value
      V6_sensor2 = random(100);
      V4_analog_pin = analogRead(A0);
      timeStrored=millis();
  }

  if (V7_variable!=V7_variable_stored) {
    V7_variable_stored=V7_variable;
    Serial.println("V7_variable="+String(V7_variable));
  }


}














//================== Virtuino Library (included) ===============
//==============================================================
//==============================================================
/* Only the next two functions are nedded to communicate with Virtuino 
 * Avoid to make changes to the code below  
 */


#define WRONG_KEY_MESSAGE  "Wrong Key"  
#define WELLCOME_MESSAGE  "Hello Virtuino"  
boolean debug=true;
//============================================================== virtuinoRun
//==============================================================

void virtuinoRun(){
    WiFiClient client = wifiServer.available();   // Check if a client has connected
    if (client) {
      if (debug) Serial.println("New client");
      while(!client.available())delay(1);    // Wait until the client sends some data
      String req = client.readStringUntil('\r');   // Read the first line of the request
      if (debug) Serial.println(req);
      //-- check request
      int pos = req.indexOf("GET /");
      if (pos!=-1){                                   // We have a GET message
        req.remove(0,pos+5);                          // Clear the characters GET / 
        int k= req.indexOf(" HTTP/");
        if (k>0) req.remove(k);                       // Clear the tail of the line

        //---the variable req contains the Virtuino command string. 
        String commandResponse = prepareResponse(&req); //Prepare the text that the server has to return to app

        String responseHttp= "HTTP/1.1 200 OK\r\n Content-Type: text/html\r\nConnection: close\r\n\r\n";
        responseHttp+=commandResponse;       // add the commandResponse to request response
        responseHttp+="\r\n\r\n";
        if (debug) Serial.println("ResponseHttp="+responseHttp);
        client.print(responseHttp);
        client.print("status:0");
        delay(1);
        if (debug) Serial.println("Client disonnected");
        client.flush();
    }
    else {
      Serial.println("invalid request");
      client.stop();
    }
}
}

//============================================================== prepareResponse
//==============================================================
String prepareResponse(String* commands){
   
   //---check the KEY
   #ifdef VIRTUINO_KEY 
      String virtuinoKey="";
      int keyPosStart = commands->indexOf("&KEY=");
      if (keyPosStart!=-1) {
        int keyPosEnd = commands->indexOf('&',keyPosStart+5);
        if (keyPosEnd!=-1) virtuinoKey= commands->substring(keyPosStart+5,keyPosEnd);
        else virtuinoKey= commands->substring(keyPosStart);
      }
      if (!virtuinoKey.equals(VIRTUINO_KEY)) return WRONG_KEY_MESSAGE;
   #endif
   
  //--- read all virtuino commands from request
  String commandResponse="";
  boolean checkNext=true;
  int cStart=-1;
  while (checkNext){
    cStart=commands->indexOf('&',cStart+1);
    if (cStart!=-1) {
      int cEnd=commands->indexOf('&',cStart+1);
      String oneCommand="";
      if (cEnd<0) oneCommand=commands->substring(cStart+1); 
      else oneCommand=commands->substring(cStart+1,cEnd); 
      if (debug) Serial.println("cmd: "+oneCommand);
      char commandType = oneCommand.charAt(0);
      if (commandType=='C') return "&CNT="+String(WELLCOME_MESSAGE);               // Virtuino sends the command &C=? to check the connection status   
      if ((commandType=='V') || (commandType=='T')) {
        int equalCharPosition = oneCommand.indexOf('=');
        int variableIndex  = oneCommand.substring(1,equalCharPosition).toInt(); 
        String valueString = oneCommand.substring(equalCharPosition+1); 
        String returnedValue = onCommandReceived(variableIndex,valueString);  
        if (returnedValue.length()>0) {
          commandResponse+='&';
          commandResponse+=commandType;
          commandResponse+=String(variableIndex);
          commandResponse+='=';
          commandResponse+=returnedValue;
        }
      }
    }
    else checkNext=false;   // the last command has checked
  }
  if (debug) Serial.println("Response="+commandResponse );
  if (commandResponse.length()==0) commandResponse="OK";
  return commandResponse;
}


  //=================== ENCODE - DECODE Library (included) ==========
  //=================================================================
  
  //================================================================ urldecode
String urldecode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    for (int i =0; i < str->length(); i++){
        c=str->charAt(i);
      if (c == '+'){
        encodedString+=' ';  
      }else if (c == '%') {
        i++;
        code0=str->charAt(i);
        i++;
        code1=str->charAt(i);
        c = (h2int(code0) << 4) | h2int(code1);
        encodedString+=c;
      } else{
         encodedString+=c;  
      }
      yield();
    }
    
   return encodedString;
}
//================================================================ urlencode
String urlencode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    char code2;
    for (int i =0; i < str->length(); i++){
      c=str->charAt(i);
      if (c == ' '){
        encodedString+= '+';
      } else if (isalnum(c)){
        encodedString+=c;
      } else{
        code1=(c & 0xf)+'0';
        if ((c & 0xf) >9){
            code1=(c & 0xf) - 10 + 'A';
        }
        c=(c>>4)&0xf;
        code0=c+'0';
        if (c > 9){
            code0=c - 10 + 'A';
        }
        code2='\0';
        encodedString+='%';
        encodedString+=code0;
        encodedString+=code1;
        //encodedString+=code2;
      }
      yield();
    }
    return encodedString;
}
//================================================================ h2int
unsigned char h2int(char c){
    if (c >= '0' && c <='9'){
        return((unsigned char)c - '0');
    }
    if (c >= 'a' && c <='f'){
        return((unsigned char)c - 'a' + 10);
    }
    if (c >= 'A' && c <='F'){
        return((unsigned char)c - 'A' + 10);
    }
    return(0);
}



